package RFC::Furnace;
#$Id: Furnace.pm,v 1.106 2023/11/30 12:31:26 sqko Exp $
###################################################################
#
#
#
#     Package RFC::Furnace;
#
# Simple web interface to RFCcontrol rigs used as normal furnaces
# where usually only the temperature setpoint is changed after 
# initial setup.
#
# To obtain an instance call the constructor in one of 2 ways:
#
# $id = RFC::Furnace->new($rignr);
#
# Where $rignr is an integer.
#
# Or:
#
# $id = RFC::Furnace->new($ref);
#
# Where $ref is a RFC::Rig instance.
#
#
#
# $id->testdir([$])        Returns the directory path to the specified furance run
#                          If called without arguments, returns the current furnace
#                          run directory.
# $id->runnr()             Returns the current furnace run.
# $id->list_runs()         Returns a ordered list of valid run numbers for this furnace. 
# $id->writedata()         makes a data read and stores it in the current run directory.
# $id->get_plot([$][$])    Returns an array with the filename for a png of the data for the 
#                          specified runnr and the total runtime for that run, if no runnr 
#                          is specified, a plot for current run is made.
#                          If the plot of a run has already been made, the path and filename
#                          for that plot is simply return ed unless a second argument is 
#                          specified.
# $id->get_run_info($)     Returns a hash with the main information about the specified furnace run as
#                          determined by parsing the info file.
# $id->get_runtime([$])    Returns the total runtime (in hours) for the specified furnace run.
#                          If no runnumber is specified the current furnace run is used and if
#                          that run is still running, the currently elapsed time is returned.
# $id->set_info(@)         Appends the argument to the info file for current run.
# $id->get_info([$])       Returns the content of the info file for specified furnace run.
#                          Default is current run.
# $id->get_program([$])    Returns the content of the program file for specified furnace run.
#                          Default is current run.
# $id->proglog($)          Appends the argument to the proglog file for current furnace run.
# $id->furnacelog($)       Appends to the furnace log (specific log for furnace runs).
#                          Note: does NOT append to the rig proglog file as proglog() does.
# $id->get_proglog([$])    Returns the content of the proglog file for specified furnace run.
#                          Default is current run.
# $id->get_furnacelog([$]) Returns the content of the furnacelog file for specified furnace run.
#                          Default is current run.
# $id->start_program($@)   Starts a new furnace run.
#                          The first argument must be the uername of the user starting the program.
#                          Appends the remaining argumenta to the program file for current furnace run.
# $id->stop_program([@])   Terminates the current program by removing the semafore.
#                          Any arguments passed the stop_program function is passed along to
#                          the proglog() function.
# $id->is_running()        Returns true if a program is already running for this furnace.
# $id->readstring()        performs a data aquisition and returns the resulting string.
#                          Note does NOT store the data, use writedata instead.
# $id->append_program(@)   Appends one or more lines to the current program. Note that each 
#                          instance of RFC::Furnace starts out with an empty program!
# $id->program_to_file()   Returns a string which can be directly saved to a file which
#                          can be run by /usr/local/bin/RFCfurnace/run_furnace.pl
# $id->program_to_data()   Returns a list of data points for the current program which can
#                          be used for creating a graph of the current program beeing made.
# $is->get_atemp()         Returns the current temperature as reported by the temperature controler
# $id->temp                Gets or sets the active temperature setpoint of the RFC::TempControl
#                          device. See TempControl.pm for details of the get_temp and set_temp 
#                          functions
# $id->temp                Gets or sets the active temperature ramprate of the RFC::TempControl
#                          device. See TempControl.pm for details of the get_ramp and set_ramp 
#                          functions
# $id->minramp()           Returns the minimum allowed ramprate.
# $id->maxramp()           Returns the maximum allowed ramprate.
# $id->mintemp()           Returns the minimum allowed temperature.
# $id->maxtemp()           Returns the maximum allowed temperature.
# $id->get_lines()         Returns an array with 4 elements:
#                          The number of comancs (new setpoints commands counts as does hold/wait commands, 
#                          but ramp rate commands don't).
#                          The maximum temperature setpoint.
#                          The ramprate leading up to the highest temperature setpoint.
#                          The hold time at the highest temperature setpoint.
#                          Thus a simple program (ramping to a temperature holding there and ramping down again
#                          will report 3 lines.
# $id->list_elements()     Returns a list of chemical elements which the furnace has been recorded as having
#                          been exposed to. The list is in the form of the atomic numbers, 1 for hydrogen and so on.
# $id->add_element($)      Adds the specified element to the list of seen elements. The function accepts either 
#                          the atomic number or the short name ('10' or 'Ne' for neon for instance). 
#                          The function returns 1 if the element was added to the list (that is that it has not 
#                          previously been on the list) nad 0 otherwise.
#                          Thus if you try to add the same element twice, the first call to add_element will 
#                          return 1 and the second 0.
# $id->del_element($)      removes the specified element to the list of seen elements. The function accepts either 
#                          the atomic number or the short name ('10' or 'Ne' for neon for instance). 
#                          The function returns 1 if the element was removed from the list 0 otherwise.
# $id->list_tubes()        Returns a list of allowble tube ids. If an empty list is returned the furnace is assumed 
#                          to be a chamber furnace. The allowed_tubes key in the main section in the rig's 
#                          configuration contain the list of allowed tubes and can only be modified by the 
#                          certified users.
#
# RFC::Furnace also implements the Visitor pattern by suppliying the Accept() method.
#
# $id->Accept($)      Runs the Accept() method on the underlying RFC::Rig instance as well
#                     as calling the VisitComplex() function on the visitor.
# $id->ClearVisitor() Special function which clears the list of already 
#                     seen visitors
#
#
#
# RFC::Furnace also has a few class functions which can be called without an RFC::Furnace instance. 
# These function are described below.
#
# RFC::Furnace::can_start_furnace() : This function returns true if the client system is allowed to start and stop
#                                     furnace runs.
#                                     the host_allow key in the RFCfurnace section in the RFCcontrol configuration
#                                     file contains the list of approved IP-addresses.
#                                     If the list is empty, all clients can start and stop furnaces.
#                                     Note  that the local system can always start and stop furnaces.
#
# RFC::Furnace::is_authorised()     : This funciton determines if a client system can access the ajax_furnace.cgi
#                                     interface without user authentification (used mainly for reporting purposes).
#                                     The host_certify key in the RFCfurnace section determines which clients are
#                                     allowed to do this. Only clients on this list can do so and if the list is
#                                     empty no-one can.
#                                     Note  that the local system can always access the ajax_furnace.cgi script
#                                     without user authentification, as any user calling this script could just as
#                                     easily access the data normally.
#
# RFC::Furnace::get_element_data()  : This function returns an array of hashes cotaining element information used
#                                     for the page showing the periodic table of elements used for selecting sample
#                                     composition.
# 
#
# RFC::Furnace::get_tube_elements($): Returns a list ov elements which a given furnace tube has seen on this server
#                                     Is intended to be used for allowing the sharing of furnace tubes between furnaces
#                                     and servers.
#                                     This function should be called for all servers and compiled to a single list before
#                                     displaying the result to the user.
# 
# RFC::Furnace::add_tube_element($$): Appends a specific element to the list of elements a furnace tube has sen on 
#                                     this server (similar to add_element() for furnaces).
#                                     Arguments: tube, element.
#
# RFC::Furnace::del_tube_element($$): Removes a specific element from the list of elements a furnace tube has sen on 
#                                     this server (similar to del_element() for furnaces).
#                                     Arguments: tube, element.
#
###################################################################


use strict;
use warnings;
use Debug;
use SemaforeFile;
use RFC::Header 7.0.0;
use RFC::Main;
use CGI qw(:standard :netscape);
use RFC::Rig;
use File::Touch;
use Date::Parse;
use Sys::Hostname qw(hostname);
use Socket;
our @ISA = qw(Debug);

use overload
    '""' => \&Rig_to_string,
    '.' => \&Rig_concat,
    'bool' => sub {return 1;},
    'fallback' => 1;

my %insth = ();
my @elements = ();
our $AUTOLOAD;

BEGIN {
    use Exporter ();
    our ($VERSION, $DISTVER, @ISA, @EXPORT, @EXPORT_OK, %EXPORT_TAGS);
    $VERSION = '$Revision: 1.106 $';
    $DISTVER = '1.2.2';
    %EXPORT_TAGS = (
        'standard'=>[],
	);
    Exporter::export_ok_tags('standard');
    Exporter::export_tags('standard');
    @ISA = qw(Exporter);
}


### Global font and line with definitions for gnuplot. Can be changed by setting to different values ###
our $lw = 'lw 2';
our $font = "font ',14'";

my $ip_addr = inet_ntoa( (gethostbyname(hostname()))[4] );

sub new
{
    my $class = shift;
    my $n = shift;
    my $rigref = undef;
    if (ref($n) eq 'RFC::Rig')
    {
	$rigref = $n;
	$n = $rigref->name();
    }
    if ($n =~ /(\d+)/)
    {
        $n = $1;
    }
    else
    {
        die "$n is not a valid furnace name (only integers allowed)\n";
    }
    my $nstr = 'Rig '.$n;
    ## Insure singleton for a given rig number (only one RFC::Rig instance for each rig ##
    foreach my $key (keys %RFC::Furnace::insth)
    {
        next unless ($key eq $nstr); 
        my $self = $RFC::Furnace::insth{$key};
        $self->writedebug("Returning existing instance of $self") if ($self->debug);
        return $self;
    }
    my $self = $class->SUPER::new();
    $self->{'_name'} = $n;
    bless $self,$class;
    if ($rigref)
    {
	$self->{'_rig'} = $rigref;
    }
    else
    {
	my $r = RFC::Rig->new($n);
	my $error = $r->error();
	if (!$r || $error)
	{
	    my $str = "Could not correctly initialise RFCcontrol rig $n!";
	    $str .= "\nError: $error\n" if ($error);
	    die $str,"\n";
	}
	$self->{'_rig'} = $r;
    }
    $RFC::Rig::insth{$self->Rig_to_string} = $self;
    my $wdir = $RFC::Header::publicdir;
    $self->{'_runfile'} = SemaforeFile->new($wdir.'rig'.$n.'/current_program.txt','/home/celltest/Furnace.lock');
    $self->{'_runnr'} = SemaforeFile->new($wdir.'rig'.$n.'/furnace_run_nr.txt','/home/celltest/Furnace.lock');
    unless ($self->{'_runnr'}->exist)
    {
	$self->{'_runnr'}->lock_ex;
	$self->{'_runnr'}->writeline(0);
	chmod 0666,$self->{'_runnr'}->filename;
	$self->{'_runnr'}->lock_un;
    }
    my $datadir = $wdir.'rig'.$n.'/furnace_runs/';
    unless (-d $datadir)
    {
	mkdir $datadir;
	## Use system as perl's chmod does not handle sticky bits well... ##
	system("/usr/bin/chmod 2775 $datadir");
    }
    $self->{'_datadir'} = $datadir;
    $self->{'_datafile'} = undef;
    $self->{'_program'} = [];
    ### Check for one and only one controlable furnace ###
    my @devs = $self->rig->list_devices_control('tempcontrol');
    die "RFC::Furnace only works for RFCcontrol rigs with exactly one (1) active temperature controler!\n" unless (@devs == 1);
    $self->{'_dev'} = $self->rig->get_device('tempcontrol',$devs[0]);

    return $self;
}

sub temp()
{
    my $self = shift;
    if (@_)
    {
	my $temp = shift;
	my $user = param('name');
	my $str = "Setting temperature directly to $temp C";
	if ($user)
	{
	    $self->proglog($str,$user);
	}
	else
	{
	    $self->proglog($str);
	}
	return $self->{'_dev'}->set_temp($temp,$user,'nolog');
    }
    else
    {
	return $self->{'_dev'}->get_temp();
    }
}

sub minramp()
{
    my $self = shift;
    return $self->{'_dev'}->minramp();
}

sub maxramp()
{
    my $self = shift;
    return $self->{'_dev'}->maxramp();
}

sub mintemp()
{
    my $self = shift;
    return $self->{'_dev'}->mintemp();
}

sub maxtemp()
{
    my $self = shift;
    return $self->{'_dev'}->maxtemp();
}


sub ramp()
{
    my $self = shift;
    if (@_)
    {
	my $temp = shift;
	my $user = param('name');
	my $str = "Setting ramprate directly to $temp C/hour";
	if ($user)
	{
	    $self->proglog($str,$user);
	}
	else
	{
	    $self->proglog($str);
	}
	return $self->{'_dev'}->set_ramp($temp,$user,'nolog');
    }
    else
    {
	return $self->{'_dev'}->get_ramp();
    }
}

sub get_atemp
{
    my $self = shift;
    return $self->{'_dev'}->get_atemp(@_);
}

sub append_program($)
{
    my $self = shift;
    while (my $line = shift)
    {
	push @{$self->{'_program'}},$line;
    }
}

sub program_to_file()
{
    my $self = shift;
    return join("\n",$self->output_program());
}

sub output_program()
{
    my $self = shift;
    my $t = $self->{'_dev'}->get_atemp();
    my $r = $self->{'_dev'}->get_ramp();
    my $n = $self->{'_dev'}->name;
    my $line = 'Starttemp = '.$t;
    my @list = ();
    my $totalt = 0;
    foreach my $str (@{$self->{'_program'}})
    {
	next unless ($str && $str =~ /\w/);
	next if ($str =~ /rampwait/i);
	$str =~ s/\,/\./g;
	if ($str =~ /Starttemp\s*\=\s*([\d\.\-\+]+)/)
	{
	    $t = $1;
	}
	if ($str =~ /^\s*ramp\s*\=\s*([\d\.\+\-]+)/)
	{
	    $r = $1;
	    push @list,"ramp = $r $n";
	    next;
	}
	if ($str =~ /^temp\s*\=\s*([\d\-\+]+)/)
	{
	    my $temp = $1;
	    push @list,"temp = $temp $n";
	    my $wait = POSIX::ceil(abs($temp - $t)*60/$r);
	    push @list,"rampwait = $wait";
	    $totalt += $wait;
	    $t = $temp;
	    next;
	}
	### Wait in minutes
	if ($str =~ /wait\s*\=\s*([\d\.]+)/)
	{
	    $totalt += $1;
	}
	### Hold in hours
	if ($str =~ /hold\s*\=\s*([\d\.]+)/)
	{
	    $totalt += $1*60;
	}
	push @list,$str;
    }
    unless (@list && $list[0] =~ /Starttemp/i)
    {
	unshift @list,$line;
    }
    push @list,sprintf("Total time = %.2f",$totalt/60); ## Total time in hours
    return @list;
}

sub program_to_data()
{
    my $self = shift;
    my $time = 0;
    my $temp = 0;
    my @res = ();
    foreach my $str ($self->output_program)
    {
	$str =~ s/\,/\./g;
	if ($str =~ /Starttemp\s*\=\s*([\d\.\+\-]+)/)
	{
	    $temp = $1;
	    push @res,$time."\t".$temp;
	    next;
	}
	if ($str =~ /^\s*addjust_temp\s+(\d+)\s+(\d+)\s*(\w*)/)
	{
	    $time += 180;
	    push @res,$time."\t".$temp;
	    next;
	}
	if ($str =~ /wait\s*\=\s*([\d\.]+)/)
	{
	    ### Wait in minutes
	    $time += $1;
	    push @res,$time."\t".$temp;
	    next;
	}
	if ($str =~ /hold\s*\=\s*([\d\.]+)/)
	{
	    ### Hold in hours
	    $time += $1 * 60;
	    push @res,$time."\t".$temp;
	    next;
	}
	if ($str =~ /temp\s*\=\s*([\d\-\+]+)/)
	{
	    $temp = $1;
	}
    }
    return @res;
}

sub testdir
{
    my $self = shift;
    my $runnr = $self->runnr();;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = 0;	
    }
    my $dir = $self->{'_datadir'}.$runnr.'/';
    unless (-d $dir)
    {
	mkdir $dir;
	## Use system as perl's chmod does not handle sticky bits well... ##
	system("/usr/bin/chmod 2775 $dir");
    }
    return $dir;
}

sub runnr
{
    my $self = shift;
    my @lines = $self->{'_runnr'}->readlines();
    return $lines[0] if (@lines);
    return 0;
}

sub list_runs()
{
    my $self = shift;
    my @list = ();
    for (my $x = 1; $x <= $self->runnr(); $x++)
    {
	push @list,$x if (-d $self->{'_datadir'}.$x);
    }
    return @list;
}

sub writedata
{
    my $self = shift;
    ### only do program datalog inside a program (baseline logging is done by noromal RFCcontrol) ###
    return undef unless ($self->is_running());
    my $data = $self->readstring();
    $self->datafile->append($data);
    return $data;
}

sub get_lines
{
    my $self = shift;
    my $count = 0;
    my $maxt = -300;
    my $ramp = 0;
    my $r = 0;
    my $hold = 0;
    my $istemp = 0;
    foreach my $str ($self->output_program())
    {
	next if ($str =~ /rampwait|starttemp/i);
	if ($str =~ /^\s*ramp\s*\=\s*([\d\.\+\-]+)/)
	{
	    $r = $1;
	    next;
	}
	if ($str =~ /temp\s*\=\s*([\d\-\+]+)/)
	{
	    $count++;
	    my $temp = $1;
	    if ($temp > $maxt)
	    {
		$maxt = $temp;
		$ramp = $r;
		$istemp = 1;
	    }
	    next;
	}
	if ($str =~ /wait\s*\=\s*([\d\.]+)/)
	{
	    ### Wait in minutes
	    my $time = $1;
	    $count++;
	    if ($istemp)
	    {
		$hold = $time / 60;
		$istemp = 0;
	    }
	    next;
	}
	if ($str =~ /hold\s*\=\s*([\d\.]+)/)
	{
	    ### Hold in hours
	    my $time = $1;
	    $count++;
	    if ($istemp)
	    {
		$hold = $time;
		$istemp = 0;
	    }
	    next;
	}	
    }
    return ($count,$maxt,$ramp,$hold);
}


sub get_plot
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    my @unlink = ();
    my $gpfn = '/tmp/RFCfurnace_plot.gp.'.$$; 
    my $mint = '';
    my $maxt = '';
    ## Get minimum maxtemp from rig config if specified ##
    my $t = $self->rig->get_cv('main','minimum_maxtemp_plot');
    $maxt = $t if ($t);
    open GP,'>',$gpfn or die "Could not open file '$gpfn' for writing: $!\n";
    push @unlink,$gpfn;
    my $png = '';
    my $dogp = 1;
    my $totaltime = 0;
    my $po2_min = $self->rig->get_cv('main','pO2_plot_min_mV');
    my $po2_max = $self->rig->get_cv('main','pO2_plot_max_mV');
    $po2_min = 0 unless ($po2_min);
    $po2_max = 1200 unless ($po2_max);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
	if ($runnr < 1)
	{
	    $runnr = 1;
	}
	if ($runnr > $self->runnr())
	{
	    $runnr = $self->runnr();
	}
	$png = $self->testdir($runnr).'profile.png';
	if (@_ || ! -e $png)
	{
	    ### Plot program and data for run ###
	    my @data = $self->get_data($runnr);
	    my $startt = 0;
	    my %cols = ();
	    my %vcols = ();
	    if (@data)
	    {
		my $head = $data[0];
		my @l = split(/[\s\t]/,$head);
		$startt = $l[2];
		for (my $x = 0; $x < @l; $x+= 4)
		{
		    next unless ($l[$x]);
		    if ($l[$x] =~ /_TEMPCONTROL_\w+/i && $l[$x] !~ /output/i)
		    {
			$l[$x] =~ /_TEMPCONTROL_(\w+)/i;
			$cols{$1} = $x+4;
			next;
		    }
		    if ($l[$x] =~ /_TEMPLOG_(\w+)/i)
		    {
			$cols{$1} = $x+4;
			next;
		    }
		    if ($l[$x] =~ /_SIMPLE_(\w+)/i)
		    {
			$vcols{$1} = $x+4;
			next;
		    }		    
		}

	    }
	    $self->append_program($self->get_program($runnr));
	    my $datafn = '/tmp/RFCfurnace_plot.data.'.$$; 
	    open MYOUT,'>',$datafn or die "Could not open file '$datafn' for writing: $!\n";
	    foreach my $line ($self->program_to_data())
	    {
		my @l = split(/\t/,$line);
		my $t = $l[1];
		$mint = $t if ($mint eq '' || $t < $mint);
		$maxt = $t if ($maxt eq '' || $t > $maxt);
		
		print MYOUT $line,"\n";
		$totaltime = $l[0];
	    }
	    close MYOUT;
	    $maxt = POSIX::ceil(($maxt + 10) / 50) * 50;
	    $mint = POSIX::floor(($mint - 10) / 50) * 50;
	    push @unlink,$datafn;
	    my $t = '';
	    my $plotcmd = "plot '$datafn' u (\$1/60):2 w lines lt rgb '#054990' $lw tit 'Program'";
	    my $dfn = $self->testdir($runnr).'data.txt';
	    foreach my $key (sort keys %cols)
	    {
		my $col = $cols{$key};
		my $tit = ucfirst($key);
		$tit =~ s/_/ /g;
		if ($tit =~ /setpoint/i)
		{
		    $plotcmd .= ",'$dfn' u ((\$3-$startt)/3600):(f(\$$col)) w lines lt rgb '#000000' $lw tit '$tit'";
		}
		else
		{
		    $plotcmd .= ",'$dfn' u ((\$3-$startt)/3600):(f(\$$col)) w lines lt rgb '#990000' $lw tit '$tit'";
		}
	    }
	    foreach my $key (sort keys %vcols)
	    {
		my $col = $vcols{$key};
		my $tit = ucfirst($key);
		$tit =~ s/_/ /g;
		$plotcmd .= ",'$dfn' u ((\$3-$startt)/3600):(g(\$$col)) axes x1y2 w lines tit '$tit'";
	    }
	    my $po2_scale = '';
	    if (%vcols)
	    {
		$po2_scale = "set y2label 'pO2 (mV)'  tc rgb 'black'\nset y2range[$po2_min:$po2_max]\nset y2tics nomirror tc rgb 'black'";
	    }
	    my $title = "$self run $runnr";
	    print GP <<END;

set out '$png'
set size 0.9,0.9
set term pngcairo background rgb 'white'
set style fill solid 1.0 noborder
set border lc rgb 'black'
set xlabel 'Time (hours)' tc rgb 'black'
set ylabel 'Temperature (C)' tc rgb 'black'
set yrange[$mint:$maxt]
$po2_scale
f(x) = x>-300?x:1/0
g(x) = x<10000?x:1/0
set key textcolor rgb 'black'
set tit '$title' tc rgb 'black'
set xtics nomirror tc rgb 'black'
set ytics nomirror tc rgb 'black'
$plotcmd

END

	}
	else
	{
	    $dogp = 0;
	}
    }
    else
    {
	my $mint = '';
	my $maxt = '';
	### Plot graph of program instead
	$png = $self->{'_datadir'}.'profile.png';
	my $datafn = '/tmp/RFCfurnace_plot.data.'.$$; 
	open MYOUT,'>',$datafn or die "Could not open file '$datafn' for writing: $!\n";
	foreach my $line ($self->program_to_data())
	{
	    my @l = split(/\t/,$line);
	    my $t = $l[1];
	    $mint = $t if ($mint eq '' || $t < $mint);
	    $maxt = $t if ($maxt eq '' || $t > $maxt);
	    print MYOUT $line,"\n";
	    $totaltime = $l[0];
	}
	close MYOUT;
	$maxt = POSIX::ceil(($maxt + 10) / 50) * 50;
	$mint = POSIX::floor(($mint - 10) / 50) * 50;
	push @unlink,$datafn;
	print GP <<END;

set out '$png'
set size 0.9,0.9
set term pngcairo background rgb 'white'
set xlabel 'Time (hours)' tc rgb 'black'
set ylabel 'Temperature (C)' tc rgb 'black'
set yrange[$mint:$maxt]
set border 3 lc rgb 'black'
set nokey
set xtics nomirror tc rgb 'black'
set ytics nomirror tc rgb 'black'
set title 'Current program profile' tc rgb 'black'
plot '$datafn' u (\$1/60):2 w lines lt rgb '#054990' $lw 

END

    }
    close GP;
    my $cmd = '/usr/bin/gnuplot '.$gpfn;
    system($cmd) if ($dogp);
    ### Cleanup ##
    foreach my $f (@unlink)
    {
	unlink $f if (-e $f);
    }
    ### Set permissions if possible ###
    eval {
	chmod 0664,$png;
    };
    ### Return path and filename to a png of the data ###
    return ($png,$totaltime);
}

sub get_runtime($)
{
    my $self = shift;
    my $runnr = shift;
    $runnr = $self->runnr() unless ($runnr);
    my @lines = $self->get_proglog($runnr);
    my $start = time;
    my $end = time;
    if (@lines)
    {
	my ($t,$str) = split(/\s+\:\s+/,$lines[0]);
	$start = str2time($t);
    }
    if ($runnr == $self->runnr() && $self->is_running())
    {
	return ($end-$start)/3600;
    }
    my ($t,$str) = split(/\s+\:\s+/,$lines[-1]);
    $end = str2time($t);
    return ($end-$start)/3600;
}


sub get_run_info($)
{
    my $self = shift;
    my $runnr = shift;
    $runnr = $self->runnr() unless ($runnr);
    my @lines = $self->get_info($runnr);
    my %res = ();
    foreach my $l (@lines)
    {
	if ($l =~ /started:\s+([\w\s\:]+)/)
	{
	    my $date = $1;
	    my ($ss,$mm,$hh,$day,$month,$year,$zone) = strptime($date);
	    $month++; ## Fix 0-based months...
	    $res{'start_date'} = sprintf("%04i-%02i-%02i %02i:%02i",($year+1900),$month,$day,$hh,$mm);
	    next;
	}
	if ($l =~ /Started\sby:\s*(\w+)/)
	{
	    $res{'started_by'} = $1;
	    next;
	}
	if ($l =~ /Ordered\sby:\s*(\w+)/)
	{
	    $res{'ordered_by'} = $1;
	    next;
	}
	if ($l =~ /Composition:\s*(.+)/)
	{
	    $res{'composition'} = $1;
	    next;
	}
	if ($l =~ /Project:\s*(.+)/)
	{
	    $res{'project'} = $1;
	    next;
	}
    }
    my $st = shift(@lines);
    while (@lines && $st !~ /^Comments/)
    {
	$st = shift(@lines);
    }
    $st =~ s/Comments\s*\:\s*//;
    $res{'comments'} = join("\n",$st,@lines);
    return %res;
}


sub get_info
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = $self->runnr();
    }
    my $fn = $self->testdir($runnr).'Info.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    my @list = $file->readlines();
    return @list;
}


sub set_info
{
    my $self = shift;
    my $fn = $self->testdir.'Info.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    $file->append(@_);
    return 1;
}


sub get_program
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = $self->runnr();
    }
    my $fn = $self->testdir($runnr).'Program.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    my @list = $file->readlines();
    return @list;
}


sub set_program
{
    my $self = shift;
    my $fn = $self->testdir.'Program.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    $file->append(@_);
    return 1;
}

sub furnacelog
{
    my $self = shift;
    my $string = shift;
    my $l = $self->furnacelogfile->lock_ex_nb();
    unless ($l)
    {
        sleep(1);
        $l = $self->furnacelogfile->lock_ex_nb();
    }
    if ($l)
    {
# Furnace log skal nok ikke appende hvem der skriver til fil?
#	if (@_)
#	{
#	    my $user = shift;
#	    ### Assume second argument is user name and append ' by user' ###
#	    $string .= " by user '$user'";
#	}
#	else
#	{
#	    my $tty = POSIX::ttyname(1);
#	    my $user = RFC::Main::get_user_name($<);
#	    foreach ($tty,$user)
#	    {
#		$_ = '' unless (defined($_));
#	    }
#	    $string .= " by user '$user' on $tty";
#	}
	my $time = localtime();
	my $s = $time.' : '.$string;
	$self->furnacelogfile->append($s);
	$self->furnacelogfile->unlock;
	return 1;
    }
    else
    {
	warn "Could not lock proglog file for writing! Message '$string' Could not be appended!\n";
        return 0;       
    }
}


sub proglog
{
    my $self = shift;
    $self->rig->proglog(@_);
    my $string = shift;
    my $l = $self->proglogfile->lock_ex_nb();
    unless ($l)
    {
        sleep(1);
        $l = $self->proglogfile->lock_ex_nb();
    }
    if ($l)
    {
	if (@_)
	{
	    my $user = shift;
	    ### Assume second argument is user name and append ' by user' ###
	    $string .= " by user '$user'";
	}
	else
	{
	    my $tty = POSIX::ttyname(1);
	    my $user = RFC::Main::get_user_name($<);
	    foreach ($tty,$user)
	    {
		$_ = '' unless (defined($_));
	    }
	    $string .= " by user '$user' on $tty";
	}
	my $time = localtime();
	my $s = $time.' : '.$string;
	$self->proglogfile->append($s);
	$self->proglogfile->unlock;
	return 1;
    }
    else
    {
	warn "Could not lock proglog file for writing! Message '$string' Could not be appended!\n";
        return 0;       
    }
}

sub get_proglog
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = $self->runnr();
    }
    my $fn = $self->testdir($runnr).'Proglog.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    my @list = $file->readlines();
    return @list;
}

sub get_furnacelog
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = $self->runnr();
    }
    my $fn = $self->testdir($runnr).'Furnacelog.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    my @list = $file->readlines();
    return @list;
}

sub get_data
{
    my $self = shift;
    my $runnr = 0;
    $runnr = shift if (@_);
    if ($runnr && $runnr =~ /(\d+)/)
    {
	$runnr = $1;
    }
    else
    {
	$runnr = $self->runnr();
    }
    my $fn = $self->testdir($runnr).'data.txt';
    my $file = SemaforeFile->new($fn,'/home/celltest/Furnace.lock');
    my @list = $file->readlines();
    return @list;
}

sub datafile
{
    my $self = shift;
    unless ($self->{'_datafile'})
    {
	$self->{'_datafile'} = SemaforeFile->new($self->testdir.'data.txt','/home/celltest/Furnace.lock');
    }
    return $self->{'_datafile'};
}


sub proglogfile
{
    my $self = shift;
    unless ($self->{'_proglog'})
    {
	$self->{'_proglog'} = SemaforeFile->new($self->testdir.'Proglog.txt','/home/celltest/Furnace_proglog.lock');
	$self->{'_proglog'}->lock_ex();
	unless ($self->{'_proglog'}->exist)
	{
	    touch $self->{'_proglog'}->filename();
	    chmod 0664,$self->{'_proglog'}->filename();
	}
	$self->{'_proglog'}->unlock();
    }
    return $self->{'_proglog'};
}

sub furnacelogfile
{
    my $self = shift;
    unless ($self->{'_furnacelog'})
    {
	$self->{'_furnacelog'} = SemaforeFile->new($self->testdir.'Furnacelog.txt','/home/celltest/Furnace_proglog.lock');
	$self->{'_furnacelog'}->lock_ex();
	unless ($self->{'_furnacelog'}->exist)
	{
	    touch $self->{'_furnacelog'}->filename();
	    chmod 0664,$self->{'_furnacelog'}->filename();
	}
	$self->{'_furnacelog'}->unlock();
    }
    return $self->{'_furnacelog'};
}


sub readstring
{
    my $self = shift;
    return $self->rig->readstring_furnace();
}


sub stop_program
{
    my $self = shift;
    return 0 unless ($self->is_running());
    $self->proglog('Program stopped',@_);
    $self->{'_runfile'}->lock_ex();
    unlink $self->{'_runfile'}->filename if ($self->{'_runfile'}->exist);
    $self->{'_runfile'}->unlock();
    return 1;
}

sub start_program
{
    my $self = shift;
    my $user = shift;
    return undef if ($self->is_running());
    $self->{'_runnr'}->lock_ex();
    my $runnr = $self->runnr();
    $runnr++;
    $self->{'_runnr'}->writeline($runnr);
    $self->{'_runfile'}->open_write_unsafe;
    my $fh = $self->{'_runfile'}->fh();
    print $fh $runnr;
    $self->{'_runfile'}->close();
    ## Do NOT user Semafore->chmod as it tries to lock beforehand and this ###
    ## clashes with the runnr lock => deadlock!                            ###
    chmod 0666,$self->{'_runfile'}->filename;
    $self->{'_datafile'} = undef;
    $self->{'_proglog'} = undef;
    $self->{'_runnr'}->unlock();
    $user = RFC::Main::get_user_name($<) unless ($user);
    $self->proglog('Program activated',$user);
    $self->set_program(@_);
    return $runnr;
}

sub is_running
{
    my $self = shift;
    if ($self->{'_runfile'}->exist)
    {
	return 1;
    }
    return 0;
}

sub rig
{
    my $self = shift;
    return $self->{'_rig'};
}

sub Rig_concat
{
    my ($self, $other, $swap) = @_;
    my $str = $self->Rig_to_string;
    if ($swap)
    {
        return $other.$str;
    }
    else
    {
        return $str.$other;
    }
}

sub Rig_to_string
{
    my $self = shift;
    return 'Furnace '.$self->{'_name'};
}


sub list_elements()
{
    my $self = shift;
    return $self->rig->get_cv('main','seen_elements');
}


sub list_tubes()
{
    my $self = shift;
    return $self->rig->get_cv('main','allowed_tubes');
}

sub add_element($)
{
    my $self = shift;
    my $e = shift;
    return 0 unless ($e); 
    my @elements = &get_element_data(); 
    if ($e =~ /(\d+)/)
    {
	$e = $1;
    }
    elsif ($e =~ /(\w+)/)
    {
	my $str = $1;
	$e = 0;
	foreach (@elements)
	{
	    $e = $$_{'Z'} if ($$_{'short'} eq $str);
	    last if ($e);
	}
    }
    else
    {
	return 0;
    }
    my @list = $self->list_elements();
    my $not_found = 1;
    foreach (@list)
    {
	if ($_ == $e)
	{
	    $not_found = 0;
	    last;
	}
    }
    if ($not_found)
    {
	push @list,$e;
	$self->rig->change_config_value('main','seen_elements',join(',',(sort {$a <=> $b} @list)));
	my $n = $elements[$e]{'short'};
	my $str = "Element '$n' was added to the list of known elements";
	$self->proglog($str,@_);
    }
    return $not_found;
}


sub del_element($)
{
    my $self = shift;
    my $e = shift;
    return 0 unless ($e);
    my @elements = &get_element_data(); 
    if ($e =~ /(\d+)/)
    {
	$e = $1;
    }
    elsif ($e =~ /(\w+)/)
    {
	my $str = $1;
	$e = 0;
	foreach (@elements)
	{
	    $e = $$_{'Z'} if ($$_{'short'} eq $str);
	    last if ($e);
	}
    }
    else
    {
	return 0;
    }
    my @list = $self->list_elements();
    my $found = 0;
    my @list2 = ();
    foreach (@list)
    {
	if ($_ == $e)
	{
	    $found = 1;
	}
	else
	{
	    push @list2,$_;
	}
    }
    if ($found)
    {
	$self->rig->change_config_value('main','seen_elements',join(',',(sort {$a <=> $b} @list2)));
	my $n = $elements[$e]{'short'};
	my $str = "Element '$n' was removed from the list of known elements";
	$self->proglog($str,@_);
    }
    return $found;
}


sub AUTOLOAD()
{
    ################################################################
    ### AUTOLOAD() , redirect all method calls not defined by    ###
    ### RFC::Furnace to the underlying rig instance              ###
    ################################################################
    my $self = shift;
    {
        no strict "refs";
        $AUTOLOAD =~ /\:\:(\w+)$/;
        return $self->{'_rig'}->$1(@_) if ($self->{'_rig'});
    }
}

return 1;

#######################
### Class functions ###
#######################

sub can_start_furnace()
{
    my @iplist = get_cv('RFCfurnace','host_allow');
    ### If no ip addresses is specified all can start
    return 1 unless (@iplist);
    ### Get remote ip of web browser ####
    my $remip = $ENV{'REMOTE_ADDR'};
    ### Local system can always start furnaces ###
    return 1 if ($remip eq '127.0.0.1');
    return 1 if ($remip eq '::1');
    return 1 if ($remip eq $ip_addr);
    for my $str (@iplist)
    {
        return 1 if ($str eq $remip);
    }
    return 0;
}


sub is_authorised()
{
    my @iplist = get_cv('RFCfurnace','host_certify');
    ### Get remote ip of web browser ####
    my $remip = $ENV{'REMOTE_ADDR'};
    ## Local system is always authorised (can look at files directly anyways) ##
    return 1 if ($remip eq '127.0.0.1');
    return 1 if ($remip eq '::1');
    return 1 if ($remip eq $ip_addr);
    for my $str (@iplist)
    {
        return 1 if ($str eq $remip);
    }
    return 0;
}


sub get_tube_elements($)
{
    my $tube = shift;
    if ($tube && $tube =~ /(\d+)/)
    {
	my $tube = $1;
	return &RFC::Main::get_cv('furnace','tube_'.$tube.'_seen_elements');	
    }
    else
    {
	return ();
    }
}

sub del_tube_element($$)
{
    my $tube = shift;
    my $element = shift;
    if ($tube && $tube =~ /(\d+)/)
    {
	$tube = $1;
    }
    else
    {
	return undef;
    }
    my @elemets = get_element_data();
    if ($element =~ /(\d+)/)
    {
	$element = $1;
    }
    elsif ($element =~ /(\w+)/)
    {
	my $n = $1;
	foreach (@elements)
	{
	    if ($n eq $$_{'short'})
	    {
		$element = $$_{'Z'};
		last;
	    }
	}
    }
    else
    {
	return undef;
    }
    my @seen = &get_tube_elements($tube);
    my $found = 0;
    my @new = ();
    foreach (@seen)
    {
	if ($element == $_)
	{
	    $found = 1;
	}
	else
	{
	    push @new,$_;
	}
    }
    return 0 unless ($found);
    my $conf = &RFC::Main::config();
    $conf->change_config_value('furnace','tube_'.$tube.'_seen_elements',join(',',@new));
    return 1;
}

sub add_tube_element($$)
{
    my $tube = shift;
    my $element = shift;
    if ($tube && $tube =~ /(\d+)/)
    {
	$tube = $1;
    }
    else
    {
	return undef;
    }
    my @elemets = get_element_data();
    if ($element =~ /(\d+)/)
    {
	$element = $1;
    }
    elsif ($element =~ /(\w+)/)
    {
	my $n = $1;
	foreach (@elements)
	{
	    if ($n eq $$_{'short'})
	    {
		$element = $$_{'Z'};
		last;
	    }
	}
    }
    else
    {
	return undef;
    }
    my @seen = &get_tube_elements($tube);
    my $new = 1;
    foreach (@seen)
    {
	next unless ($element == $_);
	$new = 0;
	last;
    }
    return 0 unless ($new);
    my @list = sort {$a <=> $b} (@seen,$element);
    my $conf = &RFC::Main::config();
    $conf->change_config_value('furnace','tube_'.$tube.'_seen_elements',join(',',@list));
    return 1;
}


sub get_element_data()
{
    return @elements if (@elements);
    my $data = <<END;

   0 | Neutron               | _     |      0 |      0 | _                |
   1 | Hydrogen              | H     |      1 |      1 | Nonmetal         | 
   2 | Helium                | He    |     18 |      1 | Noble_gas        | 
   3 | Lithium               | Li    |      1 |      2 | Alkli_metal      | 
   4 | Beryllium             | Be    |      2 |      2 | Alkaline_metal   | 
   5 | Boron                 | B     |     13 |      2 | Metalloid        | 
   6 | Carbon                | C     |     14 |      2 | Nonmetal         | 
   7 | Nitrogen              | N     |     15 |      2 | Nonmetal         | 
   8 | Oxygen                | O     |     16 |      2 | Nonmetal         | 
   9 | Fluorine              | F     |     17 |      2 | Halogen          | 
  10 | Neon                  | Ne    |     18 |      2 | Noble_gas        | 
  11 | Sodium (Natrium)      | Na    |      1 |      3 | Alkli_metal      | 
  12 | Magnesium             | Mg    |      2 |      3 | Alkaline_metal   | 
  13 | Aluminium (Aluminum)  | Al    |     13 |      3 | Other_metal      | 
  14 | Silicon               | Si    |     14 |      3 | Metalloid        | 
  15 | Phosphorus            | P     |     15 |      3 | Nonmetal         | 
  16 | Sulfur (Sulphur)      | S     |     16 |      3 | Nonmetal         | 
  17 | Chlorine              | Cl    |     17 |      3 | Halogen          | 
  18 | Argon                 | Ar    |     18 |      3 | Noble_gas        | 
  19 | Potassium (Kalium)    | K     |      1 |      4 | Alkli_metal      | 
  20 | Calcium               | Ca    |      2 |      4 | Alkaline_metal   | 
  21 | Scandium              | Sc    |      3 |      4 | Transition_metal | 
  22 | Titanium              | Ti    |      4 |      4 | Transition_metal | 
  23 | Vanadium              | V     |      5 |      4 | Transition_metal | 
  24 | Chromium              | Cr    |      6 |      4 | Transition_metal | 
  25 | Manganese             | Mn    |      7 |      4 | Transition_metal | 
  26 | Iron (Ferrum)         | Fe    |      8 |      4 | Transition_metal | 
  27 | Cobalt                | Co    |      9 |      4 | Transition_metal | 
  28 | Nickel                | Ni    |     10 |      4 | Transition_metal | 
  29 | Copper (Cuprum)       | Cu    |     11 |      4 | Transition_metal | 
  30 | Zinc                  | Zn    |     12 |      4 | Transition_metal | 
  31 | Gallium               | Ga    |     13 |      4 | Other_metal      | 
  32 | Germanium             | Ge    |     14 |      4 | Metalloid        | 
  33 | Arsenic               | As    |     15 |      4 | Metalloid        | 
  34 | Selenium              | Se    |     16 |      4 | Nonmetal         | 
  35 | Bromine               | Br    |     17 |      4 | Halogen          | 
  36 | Krypton               | Kr    |     18 |      4 | Noble_gas        | 
  37 | Rubidium              | Rb    |      1 |      5 | Alkli_metal      | 
  38 | Strontium             | Sr    |      2 |      5 | Alkaline_metal   | 
  39 | Yttrium               | Y     |      3 |      5 | Transition_metal | 
  40 | Zirconium             | Zr    |      4 |      5 | Transition_metal | 
  41 | Niobium               | Nb    |      5 |      5 | Transition_metal | 
  42 | Molybdenum            | Mo    |      6 |      5 | Transition_metal | 
  43 | Technetium            | Tc    |      7 |      5 | Transition_metal | 
  44 | Ruthenium             | Ru    |      8 |      5 | Transition_metal | 
  45 | Rhodium               | Rh    |      9 |      5 | Transition_metal | 
  46 | Palladium             | Pd    |     10 |      5 | Transition_metal | 
  47 | Silver (Argentum)     | Ag    |     11 |      5 | Transition_metal | 
  48 | Cadmium               | Cd    |     12 |      5 | Transition_metal | 
  49 | Indium                | In    |     13 |      5 | Other_metal      | 
  50 | Tin (Stannum)         | Sn    |     14 |      5 | Other_metal      | 
  51 | Antimony (Stibium)    | Sb    |     15 |      5 | Metalloid        | 
  52 | Tellurium             | Te    |     16 |      5 | Metalloid        | 
  53 | Iodine                | I     |     17 |      5 | Halogen          | 
  54 | Xenon                 | Xe    |     18 |      5 | Noble_gas        | 
  55 | Caesium (Cesium)      | Cs    |      1 |      6 | Alkli_metal      | 
  56 | Barium                | Ba    |      2 |      6 | Alkaline_metal   | 
  57 | Lanthanum             | La    |     23 |      6 | Lanthanide       | 
  58 | Cerium                | Ce    |     24 |      6 | Lanthanide       | 
  59 | Praseodymium          | Pr    |     25 |      6 | Lanthanide       | 
  60 | Neodymium             | Nd    |     26 |      6 | Lanthanide       | 
  61 | Promethium            | Pm    |     27 |      6 | Lanthanide       | 
  62 | Samarium              | Sm    |     28 |      6 | Lanthanide       | 
  63 | Europium              | Eu    |     29 |      6 | Lanthanide       | 
  64 | Gadolinium            | Gd    |     30 |      6 | Lanthanide       | 
  65 | Terbium               | Tb    |     31 |      6 | Lanthanide       | 
  66 | Dysprosium            | Dy    |     32 |      6 | Lanthanide       | 
  67 | Holmium               | Ho    |     33 |      6 | Lanthanide       | 
  68 | Erbium                | Er    |     34 |      6 | Lanthanide       | 
  69 | Thulium               | Tm    |     35 |      6 | Lanthanide       | 
  70 | Ytterbium             | Yb    |     36 |      6 | Lanthanide       | 
  71 | Lutetium              | Lu    |     37 |      6 | Lanthanide       | 
  72 | Hafnium               | Hf    |      4 |      6 | Transition_metal | 
  73 | Tantalum              | Ta    |      5 |      6 | Transition_metal | 
  74 | Tungsten (Wolfram)    | W     |      6 |      6 | Transition_metal | 
  75 | Rhenium               | Re    |      7 |      6 | Transition_metal | 
  76 | Osmium                | Os    |      8 |      6 | Transition_metal | 
  77 | Iridium               | Ir    |      9 |      6 | Transition_metal | 
  78 | Platinum              | Pt    |     10 |      6 | Transition_metal | 
  79 | Gold (Aurum)          | Au    |     11 |      6 | Transition_metal | 
  80 | Mercury (Hydrargyrum) | Hg    |     12 |      6 | Transition_metal | 
  81 | Thallium              | Tl    |     13 |      6 | Other_metal      | 
  82 | Lead (Plumbum)        | Pb    |     14 |      6 | Other_metal      | 
  83 | Bismuth               | Bi    |     15 |      6 | Other_metal      | 
  84 | Polonium              | Po    |     16 |      6 | Metalloid        | 
  85 | Astatine              | At    |     17 |      6 | Halogen          | 
  86 | Radon                 | Rn    |     18 |      6 | Noble_gas        | 
  87 | Francium              | Fr    |      1 |      7 | Alkli_metal      | 
  88 | Radium                | Ra    |      2 |      7 | Alkaline_metal   | 
  89 | Actinium              | Ac    |     23 |      7 | Actinide         | 
  90 | Thorium               | Th    |     24 |      7 | Actinide         | 
  91 | Protactinium          | Pa    |     25 |      7 | Actinide         | 
  92 | Uranium               | U     |     26 |      7 | Actinide         | 
  93 | Neptunium             | Np    |     27 |      7 | Actinide         | 
  94 | Plutonium             | Pu    |     28 |      7 | Actinide         | 
  95 | Americium             | Am    |     29 |      7 | Actinide         | 
  96 | Curium                | Cm    |     30 |      7 | Actinide         | 
  97 | Berkelium             | Bk    |     31 |      7 | Actinide         | 
  98 | Californium           | Cf    |     32 |      7 | Actinide         | 
  99 | Einsteinium           | Es    |     33 |      7 | Actinide         | 
 100 | Fermium               | Fm    |     34 |      7 | Actinide         | 
 101 | Mendelevium           | Md    |     35 |      7 | Actinide         | 
 102 | Nobelium              | No    |     36 |      7 | Actinide         | 
 103 | Lawrencium            | Lr    |     37 |      7 | Actinide         | 
 104 | Rutherfordium         | Rf    |      4 |      7 | Transition_metal | 
 105 | Dubnium               | Db    |      5 |      7 | Transition_metal | 
 106 | Seaborgium            | Sg    |      6 |      7 | Transition_metal | 
 107 | Bohrium               | Bh    |      7 |      7 | Transition_metal | 
 108 | Hassium               | Hs    |      8 |      7 | Transition_metal | 
 109 | Meitnerium            | Mt    |      9 |      7 | Transition_metal | 
 110 | Darmstadtium          | Ds    |     10 |      7 | Transition_metal | 
 111 | Roentgenium           | Rg    |     11 |      7 | Transition_metal | 
 112 | Copernicium           | Cn    |     12 |      7 | Transition_metal | 
 113 | Nihonium              | Nh    |     13 |      7 | Other_metal      | 
 114 | Flerovium             | Fl    |     14 |      7 | Other_metal      | 
 115 | Moscovium             | Mc    |     15 |      7 | Other_metal      | 
 116 | Livermorium           | Lv    |     16 |      7 | Other_metal      | 
 117 | Tenesine              | Ts    |     17 |      7 | Halogen          | 
 118 | Oganesson             | Og    |     18 |      7 | Noble_gas        | 

END


    ### Note that the neutron is added as 'element' 0 in order to awoid ###
    ### an empty first element in the list (makes iteration more easy)  ###
    foreach my $line (split(/\n/,$data))
    {
	next unless ($line && $line =~ /\w/);
	$line =~ s/^\s+//;
	my @l = split(/\s*\|\s*/,$line);
	my %h = (
	    'Z' => $l[0],
	    'name' => $l[1],
	    'short' => $l[2],
	    'egroup' => $l[3],
	    'period' => $l[4],
	    'eclass' => $l[5]
	    );
	$elements[$l[0]] = \%h;
    }
    return @elements;
}
